/* 
 *
C code for mds S-Plus library, which implements Kruskal's MDS.
(c) B.D. Ripley, May 1995.

May be used for non-commercial purposes if the source is acknowledged.
 *
 */
#include <S.h>

static int *ord;	/* ranks of dissimilarities */
static int *ord2;	/* inverse ordering (which one is rank i?) */
static int n;		/* number of  dissimilarities */
static int nr;		/* number of data points */
static int nc;		/* # cols of  fitted configuration */
static int dimx;	/* Size of configuration array */
static double *x; 	/* configuration */
static double *d;	/* dissimilarities */
static double *y; 	/* fitted distances (in rank of d order)*/
static double *yf; 	/* isotonic regression fitted values (ditto) */

void mds_fn();
static void vmmin();
static void errmsg();

/*
Download the data.
*/
void    mds_init_data(pn, pc, pr, orde, ordee, xx)
int    *pn, *pc, *pr, *orde, *ordee;
double *xx;
{
   int     i;
   n = *pn;
   nr = *pr;
   nc = *pc;
   dimx = nr * nc;
   ord = Calloc(n, int);
   ord2 = Calloc(n, int);
   x = Calloc(dimx, double);
   d = Calloc(n, double);
   y = Calloc(n, double);
   yf = Calloc(n, double);
   if (!ord | !ord2 | !x | !d | !y | !yf) errmsg("Not enough space");
   for (i = 0; i < n; i++) ord[i] = orde[i];
   for (i = 0; i < n; i++) ord2[i] = ordee[i];
   for (i = 0; i < dimx; i++) x[i] = xx[i];
}

static void calc_dist(x)
double *x;
{
   int     r1, r2, c, index;
   double  tmp, tmp1;
   index = 0;
   for (r1 = 0; r1 < nr; r1++)
      for (r2 = r1 + 1; r2 < nr; r2++) {
	 tmp = 0.0;
	 for (c = 0; c < nc; c++) {
	    tmp1 = x[r1 + c * nr] - x[r2 + c * nr];
	    tmp += tmp1 * tmp1;
	 }
	 d[index++] = sqrt(tmp);
      }
   for (index = 0; index < n; index++) y[index] = d[ord[index]];
}

static double fminfn(x)
double *x;
{
   double  ssq;
   int     do_derivatives = 0;
   calc_dist(x);
   mds_fn(y, yf, &n, &ssq, ord2, x, &nr, &nc, 0, &do_derivatives);
   return (ssq);
}

static void fmingr(x, der)
double *x, *der;
{
   double  ssq;
   int     do_derivatives = 1;
   calc_dist(x);
   mds_fn(y, yf, &n, &ssq, ord2, x, &nr, &nc, der, &do_derivatives);
}

void    mds_dovm(val, maxit, trace, xx)
double *val, *xx;
long   *maxit, *trace;
{
   int     i;
   vmmin(dimx, x, val, (int) *maxit, (int) *trace);
   for (i = 0; i < dimx; i++) xx[i] = x[i];
}

/*
Does isotonic regression.
*/

void    mds_fn(y, yf, pn, pssq, pd, x, pr, pncol, der, do_derivatives)
double *y, *yf, *pssq;
double *der, *x;
int    *pn, *pr, *pd, *pncol, *do_derivatives;
{
   int     n = *pn, i, ip, known, u, s, r = *pr, ncol = *pncol, k;
   double  tmp, ssq, *yc, slope, tstar, sstar;

   yc = Calloc((n + 1), double);
   yc[0] = 0.0;
   tmp = 0.0;
   for (i = 0; i < n; i++) {
      tmp += y[i];
      yc[i + 1] = tmp;
   }
   known = 0;
   do {
      slope = 1.0e+200;
      for (i = known + 1; i <= n; i++) {
	 tmp = (yc[i] - yc[known]) / (i - known);
	 if (tmp < slope) {
	    slope = tmp;
	    ip = i;
	 }
      }
      for (i = known; i < ip; i++)
	 yf[i] = (yc[ip] - yc[known]) / (ip - known);
   } while ((known = ip) < n);

   sstar = 0.0;
   tstar = 0.0;
   for (i = 0; i < n; i++) {
      tmp = y[i] - yf[i];
      sstar += tmp * tmp;
      tstar += y[i] * y[i];
   }
   ssq = 100 * sqrt(sstar / tstar);
   *pssq = ssq;
   Free(yc);
   if (!(*do_derivatives))
      return;
/* get derivatives */
   for (u = 0; u < r; u++) {
      for (i = 0; i < ncol; i++) {
	 tmp = 0.0;
	 for (s = 0; s < r; s++) {
	    if (s > u)
	       k = r * u - u * (u + 1) / 2 + s - u;
	    else if (s < u)
	       k = r * s - s * (s + 1) / 2 + u - s;
	    if (s != u) {
	       k = pd[k - 1];
	       tmp += ((y[k] - yf[k]) / sstar
		       - y[k] / tstar) * (x[u + r * i] - x[s + r * i]) / y[k];
	    }
	 }
	 der[u + i * r] = tmp * ssq;
      }
   }
}

/*  From here on, code borrowed from nnet lib in Venables & Ripley */

#include <stdio.h>
#include <math.h>

static double *vect();

static void errmsg(string)
char   *string;
{
   PROBLEM "%s\n", string RECOVER(NULL_ENTRY);
}

static double *vect(n)
int     n;
{
   double *v;
   v = (double *) Calloc(n, double);
   if (!v) errmsg("allocation failure in vect()");
   return v;
}

static void free_vect(v)
double *v;
{
   Free(v);
}

static double **Lmatrix(n)
int     n;
{
   int     i;
   double **m;

   m = (double **) Calloc(n, double *);
   if (!m) errmsg("fail1 in Lmatrix()");

   for (i = 0; i < n; i++) {
      m[i] = (double *) Calloc((i + 1), double);
      if (!m[i]) errmsg("fail2 in Lmatrix()");
   }
   return m;
}

static void free_Lmatrix(m, n)
double **m;
int     n;
{
   int     i;
   for (i = n - 1; i >= 0; i--) Free(m[i]);
   Free(m);
}

typedef unsigned char boolean;

#define false 0

#define stepredn	0.2
#define acctol		0.0001
#define reltest		10.0
#define abstol 		1.0e-2
#define reltol 		1.0e-3
#define REPORT		5


/*  BFGS variable-metric method, based on Pascal code
in J.C. Nash, `Compact Numerical Methods for Computers', 2nd edition,
converted by p2c then re-crafted by B.D. Ripley */

static void vmmin(n, b,  Fmin, maxit, trace)
int n; /* dimension */
double *b; /* input vect */
double *Fmin; /* minimum value */
int  maxit; /* maximum number of BFGS updates */
int trace; /* trace switch */
{
  boolean accpoint, enough;
  double  *g, *t, *X, *c, **B;
  int count, funcount, gradcount;
  double f, gradproj;
  int i, j, ilast, iter=0;
  double s, steplength;
  double D1, D2;

  g = vect(n); t = vect(n); X = vect(n); c = vect(n);
  B = Lmatrix(n);
  f = fminfn(b);
  if(trace) {printf("initial  value %f \n",f); fflush(stdout);}
  {
    *Fmin = f;
    funcount = gradcount = 1;
    fmingr(b, g); iter++;
    ilast = gradcount;

    do {
      if (ilast == gradcount) {
	for (i = 0; i < n; i++) {
	  for (j = 0; j < i; j++)
	    B[i][j] = 0.0;
	  B[i][i] = 1.0;
	}
     }

      for (i = 0; i < n; i++) {
	X[i] = b[i];
	c[i] = g[i];
      }
      gradproj = 0.0;
      for (i = 0; i < n; i++) {
	s = 0.0;
	for (j = 0; j <= i; j++)
	  s -= B[i][j] * g[j];
	for (j = i+1; j < n; j++)
	  s -= B[j][i] * g[j];
	t[i] = s;
	gradproj += s * g[i];
	}

      if (gradproj < 0.0) { /* search direction is downhill */
	steplength = 1.0;
	accpoint = false;
	do {
	  count = 0;
	  for (i = 0; i < n; i++) {
	    b[i] = X[i] + steplength * t[i];
	    if (reltest + X[i] == reltest + b[i]) /* no change */
	      count++;
	  }
	  if (count < n) {
	    f = fminfn(b);
	    funcount++;
	    accpoint = (f <= *Fmin + gradproj * steplength * acctol);

	    if (!accpoint) {
	      steplength *= stepredn;
	    }
	  }
	} while (!(count == n || accpoint));
	enough = (f > abstol) && (f < (1.0-reltol)*(*Fmin));
	/* stop if value if small or if relative change is low */
	if(!enough) count = n;
	if (count < n) { /* making progress */
	  *Fmin = f;
	  fmingr(b, g);
	  gradcount++; iter++;
	  D1 = 0.0;
	  for (i = 0; i < n; i++) {
	    t[i] = steplength * t[i];
	    c[i] = g[i] - c[i];
	    D1 += t[i] * c[i];
	  }
	  if (D1 > 0) {
	    D2 = 0.0;
	    for (i = 0; i < n; i++) {
	      s = 0.0;
	      for (j = 0; j <= i; j++)
		s += B[i][j] * c[j];
	      for (j = i+1; j < n; j++)
		s += B[j][i] * c[j];
	      X[i] = s;
	      D2 += s * c[i];
	    }
	    D2 = 1.0 + D2 / D1;
	    for (i = 0; i < n; i++) {
	      for (j = 0; j <= i; j++)
		B[i][j] += (D2 * t[i] * t[j] - X[i] * t[j] - t[i] * X[j]) / D1;
	    }
	  } else { /* D1 < 0 */
	    ilast = gradcount;
	  }
	} else {  /* no progress */
	  if (ilast < gradcount) {
	    count = 0;
	    ilast = gradcount;
	  }
	}
      } else { /* uphill search */
	count = 0;
	if (ilast == gradcount)
	  count = n;
	else
	  ilast = gradcount;
	/* Resets unless has just been reset */
      }
      if (iter%REPORT == 0 && trace) {
	printf("iter%4d value %f\n",iter,f);
	fflush(stdout);
      }
      if (iter >= maxit) break;
      if (gradcount - ilast > 2*n) ilast = gradcount; /* periodic restart */
    } while (count != n || ilast != gradcount);
  }
  if(trace) {
     printf("final  value %f \n", *Fmin);
     if(iter < maxit) 
	printf("converged\n");
     else 
	printf("stopped after %i iterations\n", iter);
  }
  free_vect(g); free_vect(t); free_vect(X); free_vect(c);
  free_Lmatrix(B,n);
}
